; == Jede injektive Abbildung auf {1,2,...,n} ist surjektiv ==

; Da wir im Induktionsschritt die Induktionshypothese auf eine Funktion ungleich
; f anwenden, muss unsere zu beweisende Aussage „∀ n,f (...)“ lauten und nicht
; „∀ f,n (...)“.
;
; Bei „∀ f,n (...)“ wird zunächst f: ℕ ⟶ ℕ (beliebig) gewählt, ist aber danach 
; in der Induktion fest. Bei „∀ n,f (...)“ wird die Induktion über „∀ f (...)“
; geführt. Hier kann sich also die Funktion im Induktionsschritt von der
; Funktion unterscheiden, auf die die Induktionshypothese angewendet wird.
(set-goal (pf "all n,f( all i ( i<=n -> f i <= n) -> all i,j (i<=n -> j<=n -> f i=f j -> i=j) -> all m ( m<=n -> ex k ( k<=n & f k = m )) )"))
;                       └───────────┬───────────┘    └──────────────────┬───────────────────┘    └──────────────────┬───────────────────┘
;                    f ist beschränkt auf {1,...,n}               f ist injektiv                           f ist surjektiv

(ind)

; === Induktionsanfang ===
(assume "f" "Boundary" "Injectivity" "m" "m<=0")

; Urbild einer Funktion f: {0} ⟶ {0} ist für jedes Bildelement 0.
;    -> Wähle für „k“ den Term „0“
(ex-intro (pt "0"))
(split)

; Beweis: 0 ≤ 0
(normalize-goal)
(use "Truth-Axiom")

; Beweis: f(0) = m
(use "nat1<=0 -> nat2<=0 -> nat1=nat2")
(use "Boundary")
(normalize-goal)
(use "Truth-Axiom")
(use "m<=0")

; === Induktionsschluss ===
(assume "n" "InductionHypothesis" "f" "Boundary" "Injectivity" "m" "m<=Succ n")

; Fallunterscheidung in f(n+1)=n+1 und f(n+1)≠n+1
(cases (pt "f(Succ n)=Succ n"))

; Fall 1: f(n+1)=n+1
(assume "f(Succ n)=Succ n")

	; Fallunterscheidung in m=n+1 und m≠n+1
	(cases (pt "m=Succ n"))

	; Fall 1.1: f(n+1)=n+1 und m=n+1
	(assume "m=Succ n")

		; im Fall „f(n+1)=n+1 und m=n+1“ ist das Urbild von m gleich n+1
		;    -> Wähle für „k“ den Term „n+1“
		(ex-intro (pt "Succ n"))
		(split)
		
		; Beweis: n+1 ≤ n+1
		(normalize-goal)
		(use "Truth-Axiom")
		
		; Beweis: f(n+1)=m
		(use "Nat=Trans" (pt "Succ n"))
		(use "f(Succ n)=Succ n")
		(use "Nat=Symm")
		(use "m=Succ n")

	; Fall 1.2: f(n+1)=n+1 und m≠n+1
	(assume "m=Succ n -> F")

		; Aus der Induktionshypothese folgt für f:
		;    ∃ k ( k ≤ n ⋀ f(k) = m )
		;
		; Beweisen wir aus dieser Aussage zunächst das aktuelle
		; Ziel „∃ k ( k ≤ n+1 ⋀ f(k) = m )“ bevor wir beweisen,
		; dass die Induktionshypothese anwendbar ist.
		(cut (pf "ex k (k<=n & f k=m)"))

		; Annahme von „∃ k ( k ≤ n ⋀ f(k) = m )“
		(assume "Ex-Assumption")

			; Aus „∃ k ( k ≤ n ⋀ f(k) = m )“ folgt die Existenz eines 
			; k0 mit k0 ≤ n und f(k0) = m
			(by-assume-with "Ex-Assumption" "k0" "And-Assumption")
			(inst-with-to "And-Assumption" 'left "k0<=n")
			(inst-with-to "And-Assumption" 'right "f k0=m")
			(drop "And-Assumption")

			; Unter unseren Annahmen ist k0 das Urbild von m
			(ex-intro (pt "k0"))
			(split)

			; Beweis: k0 ≤ n+1
			(use "nat1<=nat2 -> nat1<=Succ nat2")
			(use "k0<=n")
			
			; Beweis: f(k0) = m
			(use "f k0=m")

		; Beweis von „∃ k ( k ≤ n ⋀ f(k) = m )“
		;
		; Beachte: Um die Induktionshypothese für f(n+1) = n+1 direkt
		; auf f anwenden zu können, habe ich die Beschränktheit von f
		; mit  „∀ i ( i ≤ n ⟶ f(i) ≤ n )“ modelliert. Hätte ich 
		; „∀ i ( f(i) ≤ n )“ gewählt, so hätte ich hier eine neue Funktion h
		; einführen müssen, die „∀ i ( f(i) ≤ n )“ erfüllt (wegen f(n+1)=n+1
		; erfüllt f nicht die Aussage „∀ i ( f(i) ≤ n )“)
		(use "InductionHypothesis")
		
		; Beweis: ∀ i ( i ≤ n ⟶ f(i) ≤ n )
		(assume "i" "i<=n")
		
		; Beweis von „∀ i ( i ≤ n ⟶ f(i) ≤ n )“ durch 
		; „f(i) ≤ n+1 ⟶ f(i) ≠ n+1 ⟶ f(i) ≤ n“
		(use "nat1<=Succ nat2 -> (nat1=Succ nat2 -> F) -> nat1<=nat2")
		
		; Beweis: f(i) ≤ n+1 für i ≤ n
		(use "Boundary")
		(use "nat1<=nat2 -> nat1<=Succ nat2")
		(use "i<=n")
		
		; Beweis: f(i) ≠ n+1 für i ≤ n (mit Hilfe der Injektivität)
		(assume "f i=Succ n")
		(use "nat1<=nat2 -> nat1=Succ nat2 -> F" (pt "i") (pt "n"))
		(use "i<=n")
		(use "Injectivity")
		(use "nat1<=nat2 -> nat1<=Succ nat2")
		(use "i<=n")
		(normalize-goal)
		(use "Truth-Axiom")
		(simp "f i=Succ n")
		(use "Nat=Symm")
		(use "f(Succ n)=Succ n")

		; Beweis der Injektivität von f (auf {1,2,...n})
		(assume "i" "j" "i<=n" "j<=n" "f i=f j")
		(use "Injectivity")
		(use "nat1<=nat2 -> nat1<=Succ nat2")
		(use "i<=n")
		(use "nat1<=nat2 -> nat1<=Succ nat2")
		(use "j<=n")
		(use "f i=f j")

		; Beweis: m ≤ n
		(use "nat1<=Succ nat2 -> (nat1=Succ nat2 -> F) -> nat1<=nat2")
		(use "m<=Succ n")
		(use "m=Succ n -> F")

; Fall 2: f(n+1)≠n+1
(assume "f(Succ n)=Succ n -> F")

	; Fallunterscheidung in m=n+1 und m≠n+1
	(cases (pt "m=f(Succ n)"))

	; Fall 2.1: f(n+1)≠n+1 und m = f(n+1)
	(assume "m=f(Succ n)")

		; unter der Annahme m=f(n+1) ist n+1 das Urbild von m
		(ex-intro (pt "Succ n"))
		(split)
		
		; Beweis: n+1 ≤ n+1 
		(normalize-goal)
		(use "Truth-Axiom")
		
		; Beweis: f(n+1) = m
		(use "Nat=Symm")
		(use "m=f(Succ n)")

	; Fall 2.2: f(n+1)≠n+1 und m≠f(n+1)
	(assume "m=f(Succ n) -> F")

		; Definition der Hilfsfunktion g, die definiert ist durch:
		;
		;                 ┌ f(n+1) ; f(x) = n+1
		; g : ℕ ⟶ ℕ : x = ┤
		;                 └ f(x)   ; f(x) ≠ n+1
		;
		(if (assoc "g" PROGRAM-CONSTANTS) (remove-program-constant "g") )

		(add-program-constant "g" (py "(nat=>nat)=>nat=>nat=>nat") t-deg-one)
		(add-computation-rule (pt "g f nat1 nat2") (pt "[if (f nat2=Succ nat1) (f(Succ nat1)) (f nat2)]"))

		; Auf g kann die Induktionshypothese anwandt werden. Zunächst Beweise ich mit Hilfe
		; der Surjektivität von g die Surjektivität von f, bevor ich beweise,
		; dass die Voraussetzungen der Induktionshypothese für g erfüllt sind.
		(cut (pf "all m (m<=n -> ex k ((k<=n) & g f n k=m))"))

		(assume "SurjectivityOfG")

			; Fallunterscheidung: m=n+1 und m≠n+1
			(cases (pt "m=Succ n"))

			; Fall 2.2.1: f(n+1)≠n+1, m≠f(n+1) und m=n+1
			(assume "m=Succ n") 

				; Nach der Definition der Hilfsfunktion g ist das Urbild von m=n+1 unter f dasselbe, wie
				; das Urbild von f(n+1) unter g
				;    -> angenomme Surjektivität von g für den speziellen Bildwert f(n+1) anwenden
				(inst-with-to "SurjectivityOfG" (pt "f(Succ n)") "SurjectivityOfG-SpecialCase")
				(drop "SurjectivityOfG")

				; Um die Annahme „SurjectivityOfG-SpecialCase“ anwenden zu können, brauchen wir die Aussage
				; „f(n+1) ≤ n“ -> Aussage annehmen und später beweisen
				(cut (pf "f(Succ n)<=n"))

				(assume "f(Succ n)<=n")

					; Aus der Annahme „SurjectivityOfG-SpecialCase“ folgt nun die Existenz eines
					; k0 mit k0 ≤ n und g(k0)=f(n+1)
					(inst-with-to "SurjectivityOfG-SpecialCase" "f(Succ n)<=n" "Ex-Assumption")
					(by-assume-with "Ex-Assumption" "k0" "And-Assumption")
					(inst-with-to "And-Assumption" 'left "k0<=n")
					(inst-with-to "And-Assumption" 'right "g k0=f(Succ n)")
					(drop "And-Assumption")
					(drop "SurjectivityOfG-SpecialCase")

					; k0 ist Urbild von m=n+1 unter f
					(ex-intro (pt "k0"))
					(split)

					; Beweis: k0 ≤ n+1
					(use "nat1<=nat2 -> nat1<=Succ nat2")
					(use "k0<=n")

					; Beweis: f(k0) = m
					;
					; Um g(k0)=f(n+1) verwerten zu können, benutze ich eine Fallunterscheidung
					; in f(k0)=n+1 und f(k0)≠n+1 (Bei g wird nach f(x)=n+1 und 
					; f(x)≠n+1 unterschieden)
					(cases (pt "f k0=Succ n"))

					; Fall 2.2.1.1: f(n+1)≠n+1, m≠f(n+1), m=n+1, f(k0)=n+1
					(assume "f k0=Succ n")

						; Beweis: f(k0)=m
						(use "Nat=Trans" (pt "Succ n"))
						(use "f k0=Succ n")
						(use "Nat=Symm")
						(use "m=Succ n")

					; Fall 2.2.1.2: f(n+1)≠n+1, m≠f(n+1), m=n+1, f(k0)≠(n+1)
					(assume "f k0=Succ n -> F")

						; Formel „g(k0) = f(n+1)“ mit Hilfe der Annahme f(k0)≠n+1 vereinfachen. Man erhält:
						; 
						; f(k0) = f(n+1)
						(simphyp-with-to "g k0=f(Succ n)" "f k0=Succ n -> F" "IfElse-Formula")
						(simphyp-with-to "IfElse-Formula" "IfElse-Elim-False" (pt "f(Succ n)") (pt "f k0") "f k0=f(Succ n)")
						(drop "g k0=f(Succ n)")
						(drop "IfElse-Formula")

						; Aus f(k0) = f(n+1) folgt wegen Injektivität von f die Gleichung k0=n+1. Es ist aber
						; k0 ≤ n, so dass wir einen Widerspruch erhalten. Da so aus den Annahmen F hergeleitet
						; werden kann, kann das aktuelle Ziel f(k0)=m mit Hilfe von „Ex falso quodlibet“ hergeleitet
						; werden.
						(use "Efq")
						(use "nat1<=nat2 -> nat1=Succ nat2 -> F" (pt "k0") (pt "n"))
						(use "k0<=n")
						(use "Injectivity")
						(use "nat1<=nat2 -> nat1<=Succ nat2")
						(use "k0<=n")
						(normalize-goal)
						(use "Truth-Axiom")
						(use "f k0=f(Succ n)")

				; Beweis: f(n+1) ≤ n
				(use "nat1<=Succ nat2 -> (nat1=Succ nat2 -> F) -> nat1<=nat2")
				(use "Boundary")
				(normalize-goal)
				(use "Truth-Axiom")
				(use "f(Succ n)=Succ n -> F")

			; Fall 2.2.2: f(n+1)≠n+1, m≠f(n+1) und m≠n+1
			(assume "m=Succ n -> F")

				; Das Urbild von m≠n+1 ist unter f und g dasselbe (unter der Annahme m≠f(n+1))
				;    -> angenomme Surjektivität von g für den speziellen Bildwert m anwenden
				(inst-with-to "SurjectivityOfG" (pt "m") "SurjectivityOfG-SpecialCase")
				(drop "SurjectivityOfG")

				; Um die Annahme „SurjectivityOfG-SpecialCase“ anwenden zu können, brauchen wir die Aussage
				; „m ≤ n“ -> Aussage annehmen und später beweisen
				(cut (pf "m<=n"))

				(assume "m<=n")

					; Aus der Annahme „SurjectivityOfG-SpecialCase“ folgt nun die Existenz eines
					; k0 mit k0 ≤ n und g(k0)=m
					(inst-with-to "SurjectivityOfG-SpecialCase" "m<=n" "Ex-Assumption")
					(by-assume-with "Ex-Assumption" "k0" "And-Assumption")
					(inst-with-to "And-Assumption" 'left "k0<=n")
					(inst-with-to "And-Assumption" 'right "g k0=m")
					(drop "And-Assumption")
					(drop "SurjectivityOfG-SpecialCase")

					; Unter den getroffenen Annahmen ist k0 das Urbild von m unter f
					(ex-intro (pt "k0"))
					(split)

					; Beweis: k0 ≤ n+1 
					(use "nat1<=nat2 -> nat1<=Succ nat2")
					(use "k0<=n")

					; Um die Formel g(k0)=m aufzulösen, wird eine Fallunterscheidung in
					; f(k0)=n+1 und f(k0)≠n+1 benötigt.
					(cases (pt "f k0=Succ n"))

						; Fall 2.2.2.1: f(n+1)≠n+1, m≠f(n+1), m≠n+1 und f(k0)=n+1
						(assume "f k0=Succ n")

							; Formel „g(k0) = m“ mit Hilfe der Annahme f(k0)=n+1 vereinfachen. Man erhält:
							; 
							; f(n+1) = m
							(simphyp-with-to "g k0=m" "f k0=Succ n" "IfElse-Formula")
							(simphyp-with-to "IfElse-Formula" "IfElse-Elim-True" (pt "f(Succ n)") (pt "f k0") "f(Succ n)=m")
							(drop "g k0=m")
							(drop "IfElse-Formula")

							; f(n+1)=m steht im Widerspruch zur Annahme m≠f(n+1)
							(use "Efq")
							(use "m=f(Succ n) -> F")
							(use "Nat=Symm")
							(use "f(Succ n)=m")

						; Fall 2.2.2.2: f(n+1)≠n+1, m≠f(n+1), m≠n+1 und f(k0)≠n+1
						(assume "f k0=Succ n -> F")

							; Formel „g(k0) = m“ mit Hilfe der Annahme f(k0)≠n+1 vereinfachen. Man erhält:
							; 
							; f(k0) = m
							(simphyp-with-to "g k0=m" "f k0=Succ n -> F" "IfElse-Formula")
							(simphyp-with-to "IfElse-Formula" "IfElse-Elim-False" (pt "f(Succ n)") (pt "f k0") "f k0=m")
							(drop "IfElse-Formula")
							(drop "g k0=m")
							
							; f(k0)=m ist genau die aktuell zu beweisende Aussage
							(use "f k0=m")

				; Beweis: m ≤ n
				(use "nat1<=Succ nat2 -> (nat1=Succ nat2 -> F) -> nat1<=nat2")
				(use "m<=Succ n")
				(use "m=Succ n -> F")

		; Die Surjektivität von g wird über die Induktionshypothese bewiesen. Dementsprechend muss
		; nachgeprüft werden, dass g alle Voraussetzungen der Induktionshypothese erfüllt.
		(use "InductionHypothesis")

		; === Beweis: g ist eine Abbildung {1,...,n} ⟶ {1,...,n} ===
		(assume "i" "i<=n")

		; Um die Formel g(k0)=m aufzulösen, wird eine Fallunterscheidung in
		; f(i)=n+1 und f(i)≠n+1 benötigt.
		(cases (pt "f i=Succ n"))

		; Fall: f(i)=n+1
		(assume "f i=Succ n")

			; Ziel g(i) ≤ n mit Hilfe von f(i)=n+1 vereinfachen. Man erhält:
			;
			;  f(n+1) ≤ n
			(simp "f i=Succ n")
			(normalize-goal)

			; Beweis: f(n+1) ≤ n
			(use "nat1<=Succ nat2 -> (nat1=Succ nat2 -> F) -> nat1<=nat2")
			(use "Boundary")
			(normalize-goal)
			(use "Truth-Axiom")
			(use "f(Succ n)=Succ n -> F")

		; Fall: f(i)≠n+1
		(assume "f i=Succ n -> F")

			; Ziel g(i) ≤ n mit Hilfe von f(i)≠n+1 vereinfachen. Man erhält:
			;
			; f(i) ≤ n
			(simp "f i=Succ n -> F")
			(normalize-goal)
			
			; Beweis: f(i) ≤ n
			(use "nat1<=Succ nat2 -> (nat1=Succ nat2 -> F) -> nat1<=nat2")
			(use "Boundary")
			(use "nat1<=nat2 -> nat1<=Succ nat2")
			(use "i<=n")
			(use "f i=Succ n -> F")

		; === Beweis: g ist eine injektive Abbildung {1,...,n} ⟶ {1,...,n} ===
		(assume "i" "j" "i<=n" "j<=n" "g i=g j")

		; Fallunterscheidung in f(i)=n+1 und f(i)≠n+1 (zum Auflösen von g(i))
		(cases (pt "f i=Succ n"))

		; Fall 1: f(i)=n+1
		(assume "f i=Succ n")

			; Fallunterscheidung in f(j)=n+1 und f(j)≠n+1 (zum Auflösen von g(j))
			(cases (pt "f j=Succ n"))

			; Fall 1.1: f(i)=n+1 und f(j)=n+1
			(assume "f j=Succ n")

				; Für f(i)=n+1=f(j) folgt i=j direkt aus der Injektivität von f
				(use "Injectivity")
				(use "nat1<=nat2 -> nat1<=Succ nat2")
				(use "i<=n")
				(use "nat1<=nat2 -> nat1<=Succ nat2")
				(use "j<=n")
				(use "Nat=Trans" (pt "Succ n"))
				(use "f i=Succ n")
				(use "Nat=Symm")
				(use "f j=Succ n")

			; Fall 1.2: f(i)=n+1 und f(j)≠n+1
			(assume "f j=Succ n -> F")

				; Vereinfachung von g(i)=g(j) zu f(n+1)=f(j)
				(simphyp-with-to "g i=g j" "f i=Succ n" "temp-Formula1")
				(simphyp-with-to "temp-Formula1" "IfElse-Elim-True" (pt "f(Succ n)") (pt "f i") "temp-Formula2")
				(simphyp-with-to "temp-Formula2" "f j=Succ n -> F" "temp-Formula3")
				(simphyp-with-to "temp-Formula3" "IfElse-Elim-False" (pt "f(Succ n)") (pt "f j") "f(Succ n)=f j")
				(drop "g i=g j")
				(drop "temp-Formula1")
				(drop "temp-Formula2")
				(drop "temp-Formula3")

				; Aus f(n+1)=f(j) folgt wegen Injektivität von f die Gleichung n+1=j,
				; die im Widerspruch zu j ≤ n steht
				(use "Efq")
				(use "nat1<=nat2 -> nat1=Succ nat2 -> F" (pt "j") (pt "n"))
				(use "j<=n")
				(use "Injectivity")
				(use "nat1<=nat2 -> nat1<=Succ nat2")
				(use "j<=n")
				(normalize-goal)
				(use "Truth-Axiom")
				(use "Nat=Symm")
				(use "f(Succ n)=f j")

		; Fall 2: f(i)≠n+1
		(assume "f i=Succ n -> F")

			; Fallunterscheidung in f(j)=n+1 und f(j)≠n+1 (zum Auflösen von g(j))
			(cases (pt "f j=Succ n"))

			; Fall 2.1: f(i)≠n+1 und f(j)=n+1
			(assume "f j=Succ n")

				; Vereinfachung von g(i)=g(j) zu f(i)=f(n+1)
				(simphyp-with-to "g i=g j" "f i=Succ n -> F" "temp-Formula1")
				(simphyp-with-to "temp-Formula1" "IfElse-Elim-False" (pt "f(Succ n)") (pt "f i") "temp-Formula2")
				(simphyp-with-to "temp-Formula2" "f j=Succ n" "temp-Formula3")
				(simphyp-with-to "temp-Formula3" "IfElse-Elim-True" (pt "f(Succ n)") (pt "f j") "f i=f(Succ n)")
				(drop "g i=g j")
				(drop "temp-Formula1")
				(drop "temp-Formula2")
				(drop "temp-Formula3")

				; Aus f(i)=f(n+1) folgt wegen Injektivität von f die Gleichung i=n+1,
				; die im Widerspruch zu i ≤ n steht
				(use "Efq")
				(use "nat1<=nat2 -> nat1=Succ nat2 -> F" (pt "i") (pt "n"))
				(use "i<=n")
				(use "Injectivity")
				(use "nat1<=nat2 -> nat1<=Succ nat2")
				(use "i<=n")
				(normalize-goal)
				(use "Truth-Axiom")
				(use "f i=f(Succ n)")

			; Fall 2.2: f(i)≠n+1 und f(j)≠n+1
			(assume "f j=Succ n -> F")

				; Vereinfachung von g(i)=g(j) zu f(i)=f(j)
				(simphyp-with-to "g i=g j" "f i=Succ n -> F" "temp-Formula1")
				(simphyp-with-to "temp-Formula1" "IfElse-Elim-False" (pt "f(Succ n)") (pt "f i") "temp-Formula2")
				(simphyp-with-to "temp-Formula2" "f j=Succ n -> F" "temp-Formula3")
				(simphyp-with-to "temp-Formula3" "IfElse-Elim-False" (pt "f(Succ n)") (pt "f j") "f i=f j")
				(drop "g i=g j")
				(drop "temp-Formula1")
				(drop "temp-Formula2")
				(drop "temp-Formula3")

				; i = j folgt aus der Injektivität von f und der oben
				; vereinfachten Gleichung f(i)=f(j)
				(use "Injectivity")
				(use "nat1<=nat2 -> nat1<=Succ nat2")
				(use "i<=n")
				(use "nat1<=nat2 -> nat1<=Succ nat2")
				(use "j<=n")
				(use "f i=f j")

;; PROOF FINISHED
