; == Funktionen zur interaktiven Beweisführung mit Minlog ==

; In der Funktion "read-all-lines" verwende ich die Funktion "read-line", welche
; im Modul "ice-9" des Interpreters guile bereitgestellt wird. Dementsprechend muss
; bei einer Verwendung anderer Scheme-Interpreter die Funktion "read-all-lines" möglicherweise
; umgeschrieben werden (je nachdem, ob dieser Scheme-Interpreter die Funktion "read-line" bereit
; stellt oder nicht)
(use-modules (ice-9 rdelim))

; Index des letzten ausgeführten Befehls
(define CURRENT_LINE_NUMBER -1)

; Liste aller Scheme-Befehle eines Beweises
(define LINES '("; no proof loaded"))

; Name des aktuell geladenen Befehls
(define CURRENT_PROOF_FILE "")

; Anzahl der Befehle, die vor der aktuellen Position über
; die Funktion "show-current-state" ausgegeben werden soll
(define NUMBER_OF_LINES_TO_SHOW_BEFORE 2)

; Anzahl der Befehle, die nach der aktuellen Position über
; die Funktion "show-current-state" ausgegeben werden soll
(define NUMBER_OF_LINES_TO_SHOW_AFTER 8)

; Anzahl der Leerzeichen, die am Anfang der Funktion
; "show-current-state" ausgegeben werden soll
(define NUMBER_OF_EMPTY_LINES 50)

; Führt den Beweis der Datei "file-name" vollständig aus.
(define (run-proof file-name)
	(set! COMMENT-FLAG #f)
	(load file-name)
	(set! COMMENT-FLAG #t)
	(dcg)
)

; Lädt einen Beweis zur interaktiven Beweisausführung
(define (load-proof file-name)
	(set! CURRENT_PROOF_FILE file-name)
	(set! CURRENT_LINE_NUMBER -1)
	(reload-current-file)
	
	(next)
)

; Führt den nächsten Befehl aus
(define (next)
	(goto (+ CURRENT_LINE_NUMBER 1))
)

; Springt zum Befehl mit der Nummer "number"
(define (goto number)
	(set! COMMENT-FLAG #f)
	
	(reload-current-file)
	
	(if (< number CURRENT_LINE_NUMBER) (set! CURRENT_LINE_NUMBER -1) )
	
	(if (<= (length LINES) number) (set! number (- (length LINES) 1)) )
	
	(do ((k (max 0 (+ CURRENT_LINE_NUMBER 1)) (+ k 1))) 
		((= k (+ number 1)))
		
		(eval-string (list-ref LINES k))
	)
	
	(set! CURRENT_LINE_NUMBER number)
	
	(show-current-state)
)

; Zeigt den aktuellen Status des Beweises an
(define (show-current-state)
	(reload-current-file)
	
	(set! COMMENT-FLAG #t)
	
	
	(do ((k 0 (+ k 1)))
		((= k NUMBER_OF_EMPTY_LINES))
		(newline)
	)
	
	(do ((k (max (- CURRENT_LINE_NUMBER NUMBER_OF_LINES_TO_SHOW_BEFORE) 0) (+ k 1)))
		((= k (min (length LINES) (+ CURRENT_LINE_NUMBER (+ NUMBER_OF_LINES_TO_SHOW_AFTER 1)))))
		
		(if (= k CURRENT_LINE_NUMBER)
			(display "--> ")
			(display "    ")
		)
		
		(do ((i 0 (+ i 1)))
			((= i (max (- (floor (inexact->exact (/ (log (max 1 (length LINES) ) ) (log 10) ) ) ) (floor (inexact->exact (/ (log (max 1 k) ) (log 10) ) ) ) ) 0) ))
			
			(display " ")
		)
		
		(display k)
		(display ": ")
		(display (list-ref LINES k))
		(newline)
	)
	
	(newline)
	(display "--------------------------------------------------------------------------------")
	(newline)
	(newline)
	
	(dcg)
)

(define scs show-current-state)

; Lädt den aktuellen Beweis erneut und führt alle bisher ausgeführten Befehle
; erneut aus 
(define (reload)
	(let ((current-line CURRENT_LINE_NUMBER))
		(set! CURRENT_LINE_NUMBER -1)
		(goto current-line)
	)
)

; Lädt den aktuellen Beweis erneut
(define (reload-current-file)
	(set! LINES (read-all-lines (open-input-file CURRENT_PROOF_FILE)))
)

; Gibt eine Liste von allen Zeilen einer Eingabe "port" vom Typ String zurück.
(define (read-all-lines port)
	(let ((result '()))
		(do ((current-line (read-line port) (read-line port) ))
			((eof-object? current-line) result)
			
			(set! current-line (string-trim-both current-line))
			
			(if (and (> (string-length current-line) 0) (not (char=? (string-ref current-line 0) #\;)) )
				(set! result (append result (list current-line)))
			)
		)
	)
)
