; Diese Datei enthält alle Funktionen und Hilfssätze, die für die Ausführung des
; Beweises notwendig sind, dass injektive Abbildungen injektive Abbildung auf
; {1,2,...,n} surjektiv sind.

; Bevor diese Datei ausgeführt werden kann, muss die Datei "nat.scm" geladen
; werden („NatLeLtTrans“ ist ein Theorem aus „nat.scm“):
(if (not (assoc "NatLeLtTrans" THEOREMS))
	(begin
		(set! COMMENT-FLAG #f)
		(libload "nat.scm")
		(set! COMMENT-FLAG #t)
	)
)

; == notwendige Funktionen zur Beweisausführung ==

; Speichert ein gerade bewiesenes Theorem unter dem Namen „theorem-name“.
; Sollte es bereits ein Theorem unter dem Namen „theorem-name“ geben, so
; wird dieses Theorem überschrieben.
(define (save-or-resave theorem-name)
	(if (assoc theorem-name THEOREMS)
		(remove-theorem theorem-name)	
	)
	
	(save theorem-name)
)

; == notwendige Variablendeklarationen ==
; 
; Variablendeklarationen nur dann durchführen, wenn die Variablen noch nicht deklariert sind:

(if (not (assoc "f" VARIABLES)) (add-var-name "f" (py "nat=>nat")) )

(if (not (assoc "i" VARIABLES)) (add-var-name "i" (py "nat")) )
(if (not (assoc "j" VARIABLES)) (add-var-name "j" (py "nat")) )
(if (not (assoc "l" VARIABLES)) (add-var-name "l" (py "nat")) )

; == Lemmatas zur den natürlichen Zahlen ==
(set-goal (pf "all n( n<n -> F )"))
	(assume "n")
	(normalize-goal)
	(assume "F")
	(use "F")
; proof finished
(save-or-resave "nat1<nat1 -> F")

(set-goal (pf "all n,m ( n<=m -> m<n -> F )"))
	(assume "n" "m" "n<=m" "m<n")
	(use "nat1<nat1 -> F" (pt "n"))
	(use "NatLeLtTrans" (pt "m") (pt "n") (pt "m"))
	(use "n<=m")
	(use "m<n")
; proof finished
(save-or-resave "nat1<=nat2 -> nat2<nat1 -> F")

(set-goal (pf "all n ( n<=0 -> n=0 )"))
	(assume "n" "n<=0")
	(use "NatLeCases" (pt "0") (pt "n"))
	(use "n<=0")
	(normalize-goal)
	(use "Efq")
	(assume "n=0")
	(use "n=0")
; proof finished
(save-or-resave "nat1<=0 -> nat1=0")

(set-goal (pf "all n,m ( n<=0 -> m<=0 -> n=m )"))
	(assume "n" "m" "n<=0" "m<=0")
	(use "Nat=Trans" (pt "0"))
	(use "nat1<=0 -> nat1=0")
	(use "n<=0")
	(use "Nat=Symm")
	(use "nat1<=0 -> nat1=0")
	(use "m<=0")
; proof finished
(save-or-resave "nat1<=0 -> nat2<=0 -> nat1=nat2")

(set-goal (pf "all n,m ( n<=m -> n<=Succ m )"))
	(assume "n" "m" "n<=m")
	(use "NatLeTrans" (pt "m"))
	(use "n<=m")
	(normalize-goal)
	(use "Truth-Axiom")
; proof finished
(save-or-resave "nat1<=nat2 -> nat1<=Succ nat2")

(set-goal (pf "all n,m ( n<=m -> (n=m -> F) -> n<m )"))
	(assume "n" "m" "n<=m" "n=m -> F")
	(use "NatLeCases" (pt "m") (pt "n"))
	(use "n<=m")
	(assume "n<m")
	(use "n<m")
	(assume "n=m")
	(use "Efq")
	(use "n=m -> F")
(use "n=m")
; proof finished
(save-or-resave "nat1<=nat2 -> (nat1=nat2 -> F) -> nat1<nat2")

(set-goal (pf "all n,m ( n<=Succ m -> (n=Succ m-> F) -> n<=m)"))
	(assume "n" "m" "n<=Succ m" "n=Succ m -> F")
	(use "NatLtSuccToLe")
	(use "nat1<=nat2 -> (nat1=nat2 -> F) -> nat1<nat2")
	(use "n<=Succ m")
	(use "n=Succ m -> F")
; proof finished
(save-or-resave "nat1<=Succ nat2 -> (nat1=Succ nat2 -> F) -> nat1<=nat2")

(set-goal (pf "all n,m (n=m -> n<=m)"))
	(assume "n" "m" "n=m")
	(simp "n=m")
	(normalize-goal)
	(use "Truth-Axiom")
; proof finished
(save-or-resave "nat1=nat2 -> nat1<=nat2")

(set-goal (pf "all n,m (n<=m -> n=Succ m -> F)"))
	(assume "n" "m" "n<=m" "n=Succ m")	
	(use "nat1<=nat2 -> nat2<nat1 -> F" (pt "n") (pt "m"))
	(use "n<=m")
	(use "NatSuccLeToLt")
	(use "nat1=nat2 -> nat1<=nat2")
	(use "Nat=Symm")
	(use "n=Succ m")
; proof finished
(save-or-resave "nat1<=nat2 -> nat1=Succ nat2 -> F")

; == Lemmatas zur Fallunterscheidung ==
(set-goal (pf "all n,m ( [if True n m] = n )"))
	(assume "n" "m")
	(normalize-goal)
	(use "Truth-Axiom")
; proof finished
(save-or-resave "IfElse-Elim-True")

(set-goal (pf "all n,m ( [if False n m] = m )"))
	(assume "n" "m")
	(normalize-goal)
	(use "Truth-Axiom")
; proof finished
(save-or-resave "IfElse-Elim-False")
